<?php

use CKPL\Pay\Configuration\ConfigurationInterface;
use CKPL\Pay\Configuration\Factory\ConfigurationFactory;
use CKPL\Pay\Definition\Payment\Builder\AmountBuilder;
use CKPL\Pay\Definition\Payment\Builder\PaymentBuilder;
use CKPL\Pay\Definition\Payment\PaymentInterface;
use CKPL\Pay\Exception\Definition\AmountException;
use CKPL\Pay\Exception\Definition\PaymentException;
use CKPL\Pay\Exception\Exception;
use CKPL\Pay\Exception\Http\HttpConflictException;
use CKPL\Pay\Pay;

/**
 * Class WC_Gateway_ConotoxiaPay
 */
class WC_Gateway_Conotoxia_Pay extends WC_Payment_Gateway
{

    /**
     * @type string
     */
    const PAYMENTS_HOST = 'https://partner.cinkciarz.pl';

    /**
     * @type string
     */
    const OIDC_HOST = 'https://login.cinkciarz.pl';

    /**
     * @type string
     */
    const SANDBOX_PAYMENTS_HOST = 'https://pay-api.ckpl.us';

    /**
     * @type string
     */
    const SANDBOX_OIDC_HOST = 'https://oidc.ckpl.us';

    /**
     * @type string
     */
    const WC_LOG_FILENAME = 'woocommerce-gateway-conotoxiapay';

    /**
     * @var WC_Logger $logger
     */
    protected $logger;

    /**
     * WC_Gateway_Conotoxia_Pay constructor.
     */
    public function __construct()
    {
        $this->id = 'conotoxia_pay';
        $this->title = __('Conotoxia Pay', CONOTOXIA_PAY);
        $this->description = __('Pay by Conotoxia Pay.', CONOTOXIA_PAY);
        $this->has_fields = false;
        $this->method_title = __('Conotoxia Pay', CONOTOXIA_PAY);
        $this->method_description = __('Official Conotoxia Pay payment gateway for WooCommerce.', CONOTOXIA_PAY);
        $this->supports = [
            'products',
            'refunds'
        ];
        $this->initPaymentIcon();

        $this->init_settings();
        if ($this->get_request_parameter('page') === 'wc-settings' && $this->get_request_parameter('section') === 'conotoxia_pay') {
            $this->handleKeysGeneration();
            $this->init_form_fields();
        }

        // Hooks
        add_action('woocommerce_update_options_payment_gateways_' . $this->id, [$this, 'process_admin_options']);
        add_action('woocommerce_api_wc_gateway_' . $this->id, [$this, 'processIncomingRequest']);

        add_filter('woocommerce_thankyou_order_received_text', [$this, 'processOrderReceived'], 10, 2);
    }

    /**
     * Called when configuration page is loading
     */
    public function init_form_fields(): void
    {
        $generationToken = (string)mt_rand();
        WC()->session->set('generate_keys', $generationToken);

        $this->form_fields = [
            'enabled' => [
                'title' => __('Enable/Disable', 'woocommerce'),
                'label' => __('Enable Conotoxia Pay payment method', CONOTOXIA_PAY),
                'type' => 'checkbox',
                'default' => 'no',
            ],
            'client_id' => [
                'title' => __('API Client identifier', CONOTOXIA_PAY),
                'type' => 'text',
                'description' => __('API Client identifier can be generated in the Merchant\'s Panel on Configuration page.', CONOTOXIA_PAY),
                'desc_tip' => true
            ],
            'client_secret' => [
                'title' => __('API Client secret', CONOTOXIA_PAY),
                'type' => 'text',
                'description' => __('API Client secret can be generated in the Merchant\'s Panel on Configuration page.', CONOTOXIA_PAY),
                'desc_tip' => true
            ],
            'point_of_sale' => [
                'title' => __('Point of Sale ID', CONOTOXIA_PAY),
                'type' => 'text',
                'description' => __('Point of Sale ID is generated by the Conotoxia Pay system in the Merchant\'s Panel (e.g. POS1234567898765432).', CONOTOXIA_PAY),
                'desc_tip' => true
            ],
            'private_key' => [
                'title' => __('Private key', CONOTOXIA_PAY),
                'type' => 'textarea',
                'description' => nl2br(
                    '<a href="?page=wc-settings&tab=checkout&section=conotoxia_pay&generate_keys=' . $generationToken . '">' . __('Generate new key.', CONOTOXIA_PAY) . '</a>'
                ),
            ],
            'public_key' => [
                'title' => __('Public key', CONOTOXIA_PAY),
                'type' => 'textarea',
                'custom_attributes' => [
                    'readonly' => true
                ]
            ],
            'sandbox' => [
                'title' => __('Sandbox Mode', CONOTOXIA_PAY),
                'description' => __('Enable sandbox mode to test payments. For more information on setting up a test account, please contact our ', CONOTOXIA_PAY) . nl2br('<a href="' . __('sales www', CONOTOXIA_PAY) . '">' . __('sales department.', CONOTOXIA_PAY) . '</a>'),
                'type' => 'checkbox',
                'default' => 'no',
            ],
            'icon' => [
                'title' => __('Payment icon', CONOTOXIA_PAY),
                'description' => __('Show a payment icon on the payment method selection screen.', CONOTOXIA_PAY),
                'type' => 'checkbox',
                'default' => 'yes',
            ],
        ];
    }

    /**
     * Handling payment creation.
     *
     * @param int $orderId
     *
     * @return array|bool
     */
    public function process_payment($orderId)
    {
        $order = wc_get_order($orderId);

        try {
            $payment = $this->createPaymentFromOrder($order);
            $createdPayment = $this->initializeConotoxiaPay()->payments()->makePayment($payment);
            $order->add_order_note(sprintf(__('Payment %s has been created in Conotoxia Pay. Waiting for customer confirmation.', CONOTOXIA_PAY), $createdPayment->getPaymentId()));
            add_post_meta($orderId, '_transaction_id', $createdPayment->getPaymentId(), true);
            WC()->cart->empty_cart();

            return [
                'result' => 'success',
                'redirect' => $createdPayment->getApproveUrl()
            ];
        } catch (CKPL\Pay\Exception\Exception $e) {
            $this->log('Payment creation problem: ' . $e->getMessage());
            wc_add_notice(__('Payment creation problem', CONOTOXIA_PAY), 'error');
            return false;
        }
    }

    /**
     * @param WC_Order $order
     *
     * @return PaymentInterface
     * @throws AmountException
     * @throws PaymentException
     */
    public function createPaymentFromOrder(WC_Order $order): PaymentInterface
    {
        $returnUrl = $this->get_return_url($order);
        return (new PaymentBuilder())
            ->setExternalPaymentId($order->get_order_number())
            ->setAmount((new AmountBuilder())
                ->setValue(number_format($order->get_total(), 2, '.', ''))
                ->setCurrency(get_woocommerce_currency())
                ->getAmount())
            ->setDescription(get_bloginfo('name') . ' ' . __('order') . ' #' . $order->get_order_number())
            ->setReturnUrl($returnUrl)
            ->setErrorUrl($returnUrl)
            ->setIntegrationPlatform($this->getShopVersion())
            ->setAcceptLanguage($this->getAcceptLanguage())
            ->setNotificationUrlParameters(['wcOrderKey'=>$order->get_order_key()])
            ->denyPayLater()
            ->getPayment();
    }

    /**
     * @param int $orderId
     * @param null $amount
     * @param string $reason
     *
     * @return bool|WP_Error
     */
    public function process_refund($orderId, $amount = null, $reason = '')
    {
        $order = wc_get_order($orderId);
        if (empty($order->get_transaction_id())) {
            return false;
        }

        try {
            $sdk = $this->initializeConotoxiaPay();
            $refund = $sdk->refunds()->createRefundBuilder()
                ->setReason(empty($reason) ? __('Refund reason', CONOTOXIA_PAY) : $reason)
                ->setValue($this->formatAmount($amount))
                ->setCurrency($order->get_currency())
                ->setExternalRefundId($order->get_order_number())
                ->setPaymentId($order->get_transaction_id())
                ->setIntegrationPlatform($this->getShopVersion())
                ->setAcceptLanguage($this->getAcceptLanguage())
                ->setNotificationUrlParameters(['wcOrderKey'=>$order->get_order_key()])
                ->getRefund();
            $createdRefund = $sdk->refunds()->makeRefund($refund);
            $order->add_meta_data('_refund', ['id' => $createdRefund->getId(), 'amount' => $this->formatAmount($amount)]);
        } catch (HttpConflictException | CKPL\Pay\Exception\Http\HttpNotFoundException $e) {
            $localeIsoCode = substr(get_bloginfo("language"), 0, 2);
            $exception = $e->getTranslatedMessage($localeIsoCode) ?? $this->getRefundCreationProblem();
            wc_add_notice($exception, 'error');
            return new WP_Error(CONOTOXIA_PAY, $exception);
        } catch (Exception $e) {
            wc_add_notice(__('Refund creation problem', CONOTOXIA_PAY), 'error');
            return $this->getRefundCreationProblem();
        }
        return true;
    }

    /**
     * Called when settings are saved on configuration page
     * @return bool
     */
    public function process_admin_options(): bool
    {
        $processed = parent::process_admin_options();

        $privateKey = $this->get_option('private_key');
        if (empty($privateKey)) {
            $this->update_option('public_key');
        } else {
            $publicKey = $this->getPublicKey($privateKey);
            if ($publicKey !== null) {
                $this->update_option('public_key', $publicKey);
            } else {
                $this->update_option('public_key');
                $this->showNotice(__('Cannot generate public key from provided private key.', CONOTOXIA_PAY), 'error');
            }
        }

        if ($this->isConfigurationCompleted()) {
            try {
                $sdk = $this->initializeConotoxiaPay();
                $sdk->merchant()->invalidateToken();
                $sdk->pickSignatureKey();
                $sdk->merchant()->sendPublicKey();
                $this->showNotice(__('Public key was successfully sent to Conotoxia Pay server.', CONOTOXIA_PAY), 'success');
            } catch (CKPL\Pay\Exception\Api\PublicKeyAlreadyExistException $e) {
                $sdk->merchant()->setPublicKeyId($e->getKid());
            } catch (Exception $e) {
                $this->showNotice(__('Cannot send public key to Conotoxia Pay server: ', CONOTOXIA_PAY) . __($e->getMessage(), CONOTOXIA_PAY), 'error');
                $this->log('Cannot send public key to Conotoxia Pay server: ' . $e->getMessage());
            }
        } else {
            $this->showNotice(__('Not all fields have been completed.', CONOTOXIA_PAY));
        }

        return $processed;
    }

    /**
     * @param string $message
     * @param string $level
     */
    private function showNotice(string $message, string $level = 'warning'): void
    {
        echo '<div class="notice notice-' . $level . '"><p><strong>' . $message . '</strong></p></div>';
    }

    public function processIncomingRequest(): void
    {
        if ($_SERVER['REQUEST_METHOD'] === 'POST') {
            $input = file_get_contents('php://input');

            try {
                $sdk = $this->initializeConotoxiaPay();
                $notification = $sdk->notification()->getNotification($input);
            } catch (CKPL\Pay\Exception\Exception $e) {
                $this->log('Notification deserialization problem: ' . $e->getMessage());
                $this->sendResponse('HTTP/1.1 400 Bad Request', 'Bad Request', 400);
                return;
            }

            if ($notification->isPaymentNotification()) {
                $this->processIncomingPayment($notification);
            } elseif ($notification->isRefundNotification()) {
                $this->processIncomingRefund($notification);
            } else {
                $this->sendResponse('HTTP/1.1 400 Bad Request', 'Unsupported notification', 400);
            }
        } else {
            $this->sendResponse('HTTP/1.1 405 Method Not Allowed', 'Method Not Allowed', 405);
        }
    }

    public function processOrderReceived(): void
    {
        $orderKey = $_REQUEST['key'] ?? null;
        $data = $_REQUEST['data'] ?? null;
        if ($orderKey === null || $data === null) {
            return;
        }
        $orderId = wc_get_order_id_by_order_key($orderKey);
        $order = wc_get_order($orderId);
        $paymentMethod = $order->get_payment_method();
        if ($paymentMethod !== $this->id) {
            return;
        }

        try {
            $decodedData = $this->initializeConotoxiaPay()->payments()->decodeReturn($data);
        } catch (CKPL\Pay\Exception\Exception   $e) {
            $this->log('Return url data deserialization problem: ' . $e->getMessage());
            return;
        }
        if ($orderId !== $decodedData->getExternalPaymentId() || $order->get_transaction_id() !== $decodedData->getPaymentId()) {
            return;
        }

        switch ($decodedData->getResult()) {
            case 'SUCCESS':
                echo __(sprintf(__('Status: Payment %s completed.', CONOTOXIA_PAY), $decodedData->getPaymentId()));
                break;
            case 'SUCCESS_WITH_PAY_LATER':
                echo __(sprintf(__('Status: Payment %s pending.', CONOTOXIA_PAY), $decodedData->getPaymentId()));
                break;
            case 'REJECTED':
                echo __(sprintf(__('Status: Payment %s rejected.', CONOTOXIA_PAY), $decodedData->getPaymentId()));
                break;
            case 'ERROR':
                echo __(sprintf(__('Status: Payment %s error.', CONOTOXIA_PAY), $decodedData->getPaymentId()));
                break;
            default:
                echo __('Status: Payment error.', CONOTOXIA_PAY);
        }
    }

    /**
     * Called when configuration page is loading or admin generates keys
     */
    private function handleKeysGeneration(): void
    {
        WC()->initialize_session();
        if ($this->get_request_parameter('generate_keys') === WC()->session->get('generate_keys')) {
            $keys = $this->generateKeyPair();
            $this->update_option('private_key', $keys['private']);
            $this->update_option('public_key', $keys['public']);
            if ($this->isConfigurationCompleted()) {
                try {
                    $sdk = $this->initializeConotoxiaPay();
                    $sdk->merchant()->invalidateToken();
                    $sdk->pickSignatureKey();
                    $sdk->merchant()->sendPublicKey();
                    $this->showNotice(__('Public key was successfully sent to Conotoxia Pay server.', CONOTOXIA_PAY), 'success');
                    $this->showNotice(__('Generated keys have been saved.', CONOTOXIA_PAY), 'success');
                } catch (CKPL\Pay\Exception\Api\PublicKeyAlreadyExistException $e) {
                    $sdk->merchant()->setPublicKeyId($e->getKid());
                    $this->showNotice(__('Generated keys have been saved.', CONOTOXIA_PAY), 'success');
                } catch (Exception $e) {
                    $this->showNotice(__('Cannot send public key to Conotoxia Pay server: ', CONOTOXIA_PAY) . __($e->getMessage(), CONOTOXIA_PAY));
                    $this->showNotice(__('Generated keys have been saved.', CONOTOXIA_PAY), 'success');
                    $this->log('Cannot send public key to Conotoxia Pay server: ' . $e->getMessage());
                }
            } else {
                $this->showNotice(__('Generated keys have been saved.', CONOTOXIA_PAY), 'success');
            }
        }
    }

    /**
     * @param string $amount
     *
     * @return string
     */
    private function formatAmount(string $amount): string
    {
        return number_format($amount, 2, '.', '');
    }

    /**
     * @return bool
     */
    private function isConfigurationCompleted(): bool
    {
        $required = ['client_id', 'client_secret', 'point_of_sale', 'private_key', 'public_key'];
        foreach ($required as $option) {
            if (empty($this->get_option($option))) {
                return false;
            }
        }
        return true;
    }

    /**
     * @param CKPL\Pay\Notification\Payment\PaymentNotificationInterface $notification
     * @return void
     */
    private function processIncomingPayment(CKPL\Pay\Notification\Payment\PaymentNotificationInterface $notification): void
    {
        $orderKey = wc_get_order_id_by_order_key($notification->getAdditionalParameters()['wcOrderKey']);
        $order = wc_get_order($orderKey);
        if (!$order) {
            $this->log(sprintf('Order %s was not found for notification of %s payment.', $notification->getExternalPaymentId(), $notification->getPaymentId()));
            $this->sendResponse('HTTP/1.1 404 Not found', 'Not found', 404);
            return;
        }

        if ($notification->isCompleted()) {
            if (!$order->has_status('pending')) {
                $this->log(sprintf('Changing order %s status from %s to processing for notification of %s payment.', $notification->getExternalPaymentId(), $order->get_status(), $notification->getPaymentId()));
            }
            $order->payment_complete($notification->getPaymentId());
            $order->add_order_note(sprintf(__('Payment %s has been completed.', CONOTOXIA_PAY), $notification->getPaymentId()));
        } elseif ($notification->isCancelled()) {
            if (!$order->has_status('pending')) {
                $this->log(sprintf('Changing order %s status from %s to failed for notification of %s payment.', $notification->getExternalPaymentId(), $order->get_status(), $notification->getPaymentId()));
            }
            $message = sprintf(__('Payment %s has been cancelled.', CONOTOXIA_PAY), $notification->getPaymentId());
            $order->add_order_note($message);
            $order->update_status('failed', $message);
        } elseif ($notification->isRejected()) {
            if (!$order->has_status('pending')) {
                $this->log(sprintf('Changing order %s status from %s to failed for notification of %s payment.', $notification->getExternalPaymentId(), $order->get_status(), $notification->getPaymentId()));
            }
            $message = sprintf(__('Payment %s has been rejected.', CONOTOXIA_PAY), $notification->getPaymentId());
            $order->add_order_note($message);
            $order->update_status('failed', $message);
        }

        $this->sendResponse('HTTP/1.1 200 OK', 'OK');
    }

    /**
     * @param CKPL\Pay\Notification\Refund\RefundNotificationInterface $notification
     * @return void
     */
    private function processIncomingRefund(CKPL\Pay\Notification\Refund\RefundNotificationInterface $notification): void
    {
        $orderKey = wc_get_order_id_by_order_key($notification->getAdditionalParameters()['wcOrderKey']);
        if(is_null($orderKey)){
            $order = wc_get_order($notification->getExternalRefundId());
            $this->log('Missing wcOrderKey in notification response body - using externalRefundId field');
        }
        else {
            $order = wc_get_order($orderKey);
        }

        if (!$order) {
            $this->log(sprintf('Order was not found for notification of %s refund.', $notification->getRefundId()));
            $this->sendResponse('HTTP/1.1 404 Not found', 'Not found', 404);
            return;
        }

        if ($notification->isCompleted()) {
            $order->add_order_note(sprintf(__('Refund %s has been completed.', CONOTOXIA_PAY), $notification->getRefundId()));
        } elseif ($notification->isProcessing()) {
            $order->add_order_note(sprintf(__('Refund %s processing.', CONOTOXIA_PAY), $notification->getRefundId()));
        } elseif ($notification->isPending()) {
            $order->add_order_note(sprintf(__('Refund %s pending.', CONOTOXIA_PAY), $notification->getRefundId()));
        } elseif ($notification->isCancelled()) {
            $order->add_order_note(sprintf(__('Refund %s cancelled.', CONOTOXIA_PAY), $notification->getRefundId()));
        }

        $this->sendResponse('HTTP/1.1 200 OK', 'OK');
    }

    private function getPaymentHost(): string
    {
        if ($this->isSandboxModeEnabled()) {
            return static::SANDBOX_PAYMENTS_HOST;
        }
        return static::PAYMENTS_HOST;
    }

    /**
     * @return bool
     */
    private function isSandboxModeEnabled(): bool
    {
        return $this->get_option('sandbox') === 'yes';
    }

    /**
     * @return bool
     */
    private function isPaymentIconEnabled(): bool
    {
        return $this->get_option('icon') === 'yes';
    }

    private function getOidcHost(): string
    {
        if ($this->isSandboxModeEnabled()) {
            return static::SANDBOX_OIDC_HOST;
        }
        return static::OIDC_HOST;
    }

    private function generateKeyPair(): array
    {
        if (!extension_loaded('openssl')) {
            $this->log('Error: Missing OpenSSL extension.');
            return ['public' => null, 'private' => null];
        }
        $keyPair = openssl_pkey_new([
            'private_key_bits' => 2048,
            'private_key_type' => OPENSSL_KEYTYPE_RSA
        ]);
        openssl_pkey_export($keyPair, $privateKey);
        $publicKey = openssl_pkey_get_details($keyPair) ['key'];
        openssl_pkey_free($keyPair);
        return ['public' => trim($publicKey), 'private' => trim($privateKey)];
    }

    /**
     * @param string $privateKey
     * @return string|null
     */
    private function getPublicKey($privateKey): ?string
    {
        if (!extension_loaded('openssl')) {
            $this->log('Error:  Missing OpenSSL extension.');
            return null;
        }
        $keyPair = openssl_pkey_get_private($privateKey);
        if ($keyPair === false) {
            return null;
        }
        $publicKey = openssl_pkey_get_details($keyPair)['key'];
        openssl_pkey_free($keyPair);
        return trim($publicKey);
    }

    private function log($message): void
    {
        if (!class_exists('WC_Logger')) {
            return;
        }

        if (empty($this->logger)) {
            if ($this->is_old_wc_version()) {
                $this->logger = new WC_Logger();
            } else {
                $this->logger = wc_get_logger();
            }
        }

        if ($this->is_old_wc_version()) {
            $this->logger->add(self::WC_LOG_FILENAME, $message, WC_Log_Levels::DEBUG);
        } else {
            $this->logger->debug($message, ['source' => self::WC_LOG_FILENAME]);
        }
    }

    private function is_old_wc_version()
    {
        return version_compare(WC_VERSION, '3.0', '<');
    }

    /**
     * @return string
     */
    private function getShopVersion(): string
    {
        global $wp_version;
        return 'WORDPRESS=' . $wp_version . ';WOO=' . WC()->version . ';PLUGIN=' . CONOTOXIA_PAY_VERSION;
    }

    /**
     * @return string|null
     */
    private function getAcceptLanguage(): ?string
    {
        return $_SERVER['HTTP_ACCEPT_LANGUAGE'];
    }


    /**
     * @return WP_Error
     */
    private function getRefundCreationProblem(): WP_Error
    {
        return new WP_Error(CONOTOXIA_PAY, __('Refund creation problem.', CONOTOXIA_PAY));
    }

    private function sendResponse(string $header, string $message, $http_response_code = null): void
    {
        header($header, true, $http_response_code);
        echo $message;
        ob_flush();
    }

    /**
     * @return Pay
     * @throws \CKPL\Pay\Exception\ConfigurationException
     */
    private function initializeConotoxiaPay(): Pay
    {
        return new Pay(ConfigurationFactory::fromArray([
            ConfigurationInterface::HOST => $this->getPaymentHost(),
            ConfigurationInterface::OIDC => $this->getOidcHost(),
            ConfigurationInterface::CLIENT_ID => $this->get_option('client_id'),
            ConfigurationInterface::CLIENT_SECRET => $this->get_option('client_secret'),
            ConfigurationInterface::POINT_OF_SALE => $this->get_option('point_of_sale'),
            ConfigurationInterface::PRIVATE_KEY => $this->get_option('private_key'),
            ConfigurationInterface::PUBLIC_KEY => $this->get_option('public_key'),
            ConfigurationInterface::STORAGE => new WC_Gateway_Conotoxia_Pay_Storage(),
        ]));
    }

    protected function initPaymentIcon(): void
    {
        if ($this->isPaymentIconEnabled()) {
            $this->icon = apply_filters(CONOTOXIA_PAY . '_icon', get_site_url() . '/wp-content/plugins/conotoxiapay/images/pay_logo.svg');
        }
    }

    private function get_request_parameter(string $key, string $default = ''): string {
        if (isset($_GET[$key])) {
            return $_GET[$key];
        }
        return $default;
    }
}
